import json
import logging
from datetime import datetime, timedelta
from typing import Any
from urllib.parse import urljoin

import tornado.escape
import tornado.web
from tornado.escape import url_escape

from pinnwand import configuration, database, defensive, error, utility

log = logging.getLogger(__name__)


class Base(tornado.web.RequestHandler):
    """Base page for all 'web' pages to inherit from. This page handles
    default methods for GET and POST but more importantly overwrites
    `write_error` to render error pages.

    It automatically converts ValidationError to a 400 error page but leaves
    other HTTPErrors alone."""

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        if status_code == 404:
            self.render(
                "error.html",
                text="That page does not exist",
                status_code=404,
                pagetitle="error",
            )
        else:
            type_, exc, traceback = kwargs["exc_info"]

            if type_ == error.ValidationError:
                self.set_status(400)
                self.render(
                    "error.html",
                    text=str(exc),
                    status_code=400,
                    pagetitle="error",
                )
            elif type_ == error.RatelimitError:
                self.set_status(429)
                self.render(
                    "error.html",
                    text=str(exc),
                    status_code=429,
                    pagetitle="error",
                )
            elif type_ == error.SpamError:
                self.set_status(451)
                self.render(
                    "error.html",
                    text=str(exc),
                    status_code=429,
                    pagetitle="error",
                )
            else:
                self.render(
                    "error.html",
                    text="unknown error",
                    status_code=500,
                    pagetitle="error",
                )

    async def get(self) -> None:
        raise tornado.web.HTTPError(404)

    async def post(self) -> None:
        raise tornado.web.HTTPError(405)


class Show(Base):
    """Show a paste on the deprecated API."""

    async def get(self, slug: str) -> None:  # type: ignore
        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        with database.session() as session:
            paste = (
                session.query(database.Paste)
                .filter(database.Paste.slug == slug)
                .first()
            )

            if not paste:
                raise tornado.web.HTTPError(404)

            if paste.exp_date < datetime.utcnow():
                session.delete(paste)
                session.commit()

                log.warn(
                    "Show.get: paste was expired, is your cronjob running?"
                )

                raise tornado.web.HTTPError(404)

            self.write(
                {
                    "paste_id": paste.slug,
                    "raw": paste.files[0].raw,
                    "fmt": paste.files[0].fmt,
                    "lexer": paste.files[0].lexer,
                    "expiry": paste.exp_date.isoformat(),
                    "filename": paste.files[0].filename,
                }
            )


class Create(Base):
    """Create a paste on the deprecated API."""

    def check_xsrf_cookie(self) -> None:
        """No XSRF cookies on the API."""
        return

    async def get(self) -> None:
        raise tornado.web.HTTPError(405)

    async def post(self) -> None:
        if defensive.ratelimit(self.request, area="create"):
            raise error.RatelimitError()

        lexer = self.get_body_argument("lexer")
        raw = self.get_body_argument("code", strip=False)
        expiry = self.get_body_argument("expiry")
        filename = self.get_body_argument("filename", None)

        if not raw or not raw.strip():
            log.info("APINew.post: a paste was submitted without content")
            raise tornado.web.HTTPError(400)

        if lexer not in utility.list_languages():
            log.info("APINew.post: a paste was submitted with an invalid lexer")
            raise tornado.web.HTTPError(400)

        if expiry not in configuration.expiries:
            log.info(
                "APINew.post: a paste was submitted with an invalid expiry"
            )
            raise tornado.web.HTTPError(400)

        paste = database.Paste(
            utility.slug_create(),
            configuration.expiries[expiry],
            "deprecated-api",
        )
        paste.files.append(database.File(paste.slug, raw, lexer, filename))

        with database.session() as session:
            session.add(paste)
            session.commit()

            req_url = self.request.full_url()
            location = paste.slug
            if filename:
                location += "#" + url_escape(filename)
            self.write(
                {
                    "paste_id": paste.slug,
                    "removal_id": paste.removal,
                    "paste_url": urljoin(req_url, f"/{location}"),
                    "raw_url": urljoin(req_url, f"/raw/{paste.files[0].slug}"),
                }
            )


class Remove(Base):
    """Remove a paste through the deprecated API."""

    def check_xsrf_cookie(self) -> None:
        """No XSRF cookies on the API."""
        return

    async def post(self) -> None:
        if defensive.ratelimit(self.request, area="delete"):
            raise error.RatelimitError()

        with database.session() as session:
            paste = (
                session.query(database.Paste)
                .filter(
                    database.Paste.removal
                    == self.get_body_argument("removal_id")
                )
                .first()
            )

            if not paste:
                self.set_status(400)
                return

            session.delete(paste)
            session.commit()

            # this is set this way because tornado tries to protect us
            # by not allowing lists to be returned, looking at this code
            # it really shouldn't be a list but we have to keep it for
            # backwards compatibility
            self.set_header("Content-Type", "application/json")
            self.write(
                json.dumps([{"paste_id": paste.slug, "status": "removed"}])
            )


class Lexer(Base):
    """List lexers through the deprecated API."""

    async def get(self) -> None:
        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        self.write(utility.list_languages())


class Expiry(Base):
    """List expiries through the deprecated API."""

    async def get(self) -> None:
        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        self.write(
            {
                name: str(timedelta(seconds=delta))
                for name, delta in configuration.expiries.items()
            }
        )



import json
import logging
from datetime import timedelta
from typing import Any
from urllib.parse import urljoin

import tornado.web

from pinnwand import configuration, database, defensive, error, utility

log = logging.getLogger(__name__)


class Base(tornado.web.RequestHandler):
    def write_error(self, status_code: int, **kwargs: Any) -> None:
        _, exc, _ = kwargs["exc_info"]
        self.write({"state": "error", "code": status_code, "message": str(exc)})


class Lexer(Base):
    async def get(self) -> None:
        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        self.write(utility.list_languages())


class Expiry(Base):
    async def get(self) -> None:
        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        self.write(
            {
                name: str(timedelta(seconds=delta))
                for name, delta in configuration.expiries.items()
            }
        )


class Paste(Base):
    def check_xsrf_cookie(self) -> None:
        return

    async def get(self) -> None:
        raise tornado.web.HTTPError(405)

    async def post(self) -> None:
        if defensive.ratelimit(self.request, area="create"):
            raise error.RatelimitError()

        try:
            data = tornado.escape.json_decode(self.request.body)
        except json.decoder.JSONDecodeError:
            raise tornado.web.HTTPError(400, "could not parse json body")

        expiry = data.get("expiry")

        if expiry not in configuration.expiries:
            log.info("Paste.post: a paste was submitted with an invalid expiry")
            raise tornado.web.HTTPError(400, "invalid expiry")

        auto_scale = data.get("long", None) is None

        files = data.get("files", [])

        if not files:
            raise tornado.web.HTTPError(400, "no files provided")

        with database.session() as session, utility.SlugContext(
            auto_scale
        ) as slug_context:
            paste = database.Paste(
                next(slug_context),
                configuration.expiries[expiry],
                "v1-api",
            )

            for file in files:
                lexer = file.get("lexer", "")
                content = file.get("content")
                filename = file.get("name")

                if lexer not in utility.list_languages():
                    raise tornado.web.HTTPError(400, "invalid lexer")

                if not content:
                    raise tornado.web.HTTPError(400, "invalid content (empty)")

                try:
                    paste.files.append(
                        database.File(
                            next(slug_context),
                            content,
                            lexer,
                            filename,
                        )
                    )
                except error.ValidationError:
                    raise tornado.web.HTTPError(
                        400, "invalid content (exceeds size limit)"
                    )

            if sum(len(f.fmt) for f in paste.files) > configuration.paste_size:
                raise tornado.web.HTTPError(
                    400, "invalid content (exceeds size limit)"
                )

            paste.files[0].slug = paste.slug

            session.add(paste)

            try:
                session.commit()
            except Exception:  # XXX be more precise
                log.warning("%r", slug_context._slugs)
                raise

            # Send the client to the paste
            url_request = self.request.full_url()
            url_paste = urljoin(url_request, f"/{paste.slug}")
            url_removal = urljoin(url_request, f"/remove/{paste.removal}")

            self.write({"link": url_paste, "removal": url_removal})



import binascii
import io
import logging
import zipfile
from datetime import datetime
from typing import Any

import docutils.core
import tornado.web

from pinnwand import configuration, database, defensive, error, path, utility

log = logging.getLogger(__name__)


class Base(tornado.web.RequestHandler):
    """Base page for all 'web' pages to inherit from. This page handles
    default methods for GET and POST but more importantly overwrites
    `write_error` to render error pages.

    It automatically converts ValidationError to a 400 error page but leaves
    other HTTPErrors alone."""

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        if status_code == 404:
            self.render(
                "error.html",
                text="That page does not exist",
                status_code=404,
                pagetitle="error",
            )
        else:
            type_, exc, traceback = kwargs["exc_info"]

            if type_ == error.ValidationError:
                self.set_status(400)
                self.render(
                    "error.html",
                    text=str(exc),
                    status_code=400,
                    pagetitle="error",
                )
            elif type_ == error.RatelimitError:
                self.set_status(429)
                self.render(
                    "error.html",
                    text=str(exc),
                    status_code=429,
                    pagetitle="error",
                )
            elif type_ == error.SpamError:
                self.set_status(451)
                self.render(
                    "error.html",
                    text=str(exc),
                    status_code=429,
                    pagetitle="error",
                )
            else:
                self.render(
                    "error.html",
                    text="unknown error",
                    status_code=500,
                    pagetitle="error",
                )

    async def get(self) -> None:
        raise tornado.web.HTTPError(404)

    async def post(self) -> None:
        raise tornado.web.HTTPError(405)


class Create(Base):
    """The index page shows the new paste page with a list of all available
    lexers from Pygments."""

    async def get(self, lexers: str = "") -> None:
        """Render the new paste form, optionally have a lexer preselected from
        the URL."""

        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        lexers_available = utility.list_languages()
        lexers_selected = [
            lexer for lexer in lexers.split("+") if lexer.strip()
        ]

        # Our default lexer is just that, text
        if not lexers_selected:
            lexers_selected = ["text"]

        # Make sure all lexers are available
        if not all(lexer in lexers_available for lexer in lexers_selected):
            log.debug("CreatePaste.get: non-existent lexer requested")
            raise tornado.web.HTTPError(404)

        await self.render(
            "create.html",
            expiries=configuration.expiries,
            lexers=lexers_selected,
            lexers_available=lexers_available,
            pagetitle="Create new paste",
            message=None,
            paste=None,
        )

    async def post(self) -> None:
        """This is a historical endpoint to create pastes, pastes are marked as
        old-web and will get a warning on top of them to remove any access to
        this route.

        pinnwand has since evolved with an API which should be used and a
        multi-file paste.

        See the 'CreateAction' for the new-style creation of pastes."""

        lexer = self.get_body_argument("lexer")
        raw = self.get_body_argument("code", strip=False)
        expiry = self.get_body_argument("expiry")

        if defensive.ratelimit(self.request, area="create"):
            raise error.RatelimitError()

        if lexer not in utility.list_languages():
            log.info("Paste.post: a paste was submitted with an invalid lexer")
            raise tornado.web.HTTPError(400)

        # Guard against empty strings
        if not raw or not raw.strip():
            return self.redirect(f"/+{lexer}")

        if expiry not in configuration.expiries:
            log.info("Paste.post: a paste was submitted with an invalid expiry")
            raise tornado.web.HTTPError(400)

        paste = database.Paste(
            utility.slug_create(),
            configuration.expiries[expiry],
            "deprecated-web",
        )
        file = database.File(paste.slug, raw, lexer)
        paste.files.append(file)

        with database.session() as session:
            session.add(paste)
            session.commit()

            # The removal cookie is set for the specific path of the paste it is
            # related to
            self.set_cookie(
                "removal", str(paste.removal), path=f"/{paste.slug}"
            )

            # Send the client to the paste
            self.redirect(f"/{paste.slug}")

    def check_xsrf_cookie(self) -> None:
        """The CSRF token check is disabled. While it would be better if it
        was on the impact is both small (someone could make a paste in
        a users name which could allow pinnwand to be used as a vector for
        exfiltration from other XSS) and some command line utilities
        POST directly to this endpoint without using the JSON endpoint."""
        return


class CreateAction(Base):
    """The create action is the 'new' way to create pastes and supports multi
    file pastes."""

    def post(self) -> None:  # type: ignore
        """POST handler for the 'web' side of things."""

        if defensive.ratelimit(self.request, area="create"):
            raise error.RatelimitError()

        expiry = self.get_body_argument("expiry")

        if expiry not in configuration.expiries:
            log.info(
                "CreateAction.post: a paste was submitted with an invalid expiry"
            )
            raise error.ValidationError()

        auto_scale = self.get_body_argument("long", None) is None

        lexers = self.get_body_arguments("lexer")
        raws = self.get_body_arguments("raw", strip=False)
        filenames = self.get_body_arguments("filename")

        if not all([lexers, raws, filenames]):
            # Prevent empty argument lists from making it through
            raise error.ValidationError()

        if not all(raw.strip() for raw in raws):
            # Prevent empty raws from making it through
            raise error.ValidationError()

        if any(len(L) != len(lexers) for L in [lexers, raws, filenames]):
            log.info("CreateAction.post: mismatching argument lists")
            raise error.ValidationError()

        if any(lexer not in utility.list_languages() for lexer in lexers):
            log.info("CreateAction.post: a file had an invalid lexer")
            raise error.ValidationError()

        with database.session() as session, utility.SlugContext(
            auto_scale
        ) as slug_context:
            paste = database.Paste(
                next(slug_context), configuration.expiries[expiry], "web"
            )

            for (lexer, raw, filename) in zip(lexers, raws, filenames):
                paste.files.append(
                    database.File(
                        next(slug_context),
                        raw,
                        lexer,
                        filename if filename else None,
                    )
                )

            if sum(len(f.fmt) for f in paste.files) > configuration.paste_size:
                log.info("CreateAction.post: sum of files was too large")
                raise error.ValidationError()

            # For the first file we will always use the same slug as the paste,
            # since slugs are generated to be unique over both pastes and files
            # this can be done safely.
            paste.files[0].slug = paste.slug

            session.add(paste)
            session.commit()

            # The removal cookie is set for the specific path of the paste it is
            # related to
            self.set_cookie(
                "removal", str(paste.removal), path=f"/{paste.slug}"
            )

            # Send the client to the paste
            self.redirect(f"/{paste.slug}")


class Repaste(Base):
    """Repaste is a specific case of the paste page. It only works for pre-
    existing pastes and will prefill the textarea and lexer."""

    async def get(self, slug: str) -> None:  # type: ignore
        """Render the new paste form, optionally have a lexer preselected from
        the URL."""

        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        with database.session() as session:
            paste = (
                session.query(database.Paste)
                .filter(database.Paste.slug == slug)
                .first()
            )

            if not paste:
                raise tornado.web.HTTPError(404)

            lexers_available = utility.list_languages()

            await self.render(
                "create.html",
                expiries=configuration.expiries,
                lexers=["text"],  # XXX make this majority of file lexers?
                lexers_available=lexers_available,
                pagetitle="repaste",
                message=None,
                paste=paste,
            )


class Show(Base):
    """Show a paste."""

    async def get(self, slug: str) -> None:  # type: ignore
        """Fetch paste from database by slug and render the paste."""

        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        with database.session() as session:
            paste = (
                session.query(database.Paste)
                .filter(database.Paste.slug == slug)
                .first()
            )

            if not paste:
                raise tornado.web.HTTPError(404)

            if paste.exp_date < datetime.utcnow():
                session.delete(paste)
                session.commit()

                log.warn(
                    "Show.get: paste was expired, is your cronjob running?"
                )

                raise tornado.web.HTTPError(404)

            can_delete = self.get_cookie("removal") == str(paste.removal)

            self.render(
                "show.html",
                paste=paste,
                pagetitle=f"View paste {paste.slug}",
                can_delete=can_delete,
                linenos=False,
            )


class RedirectShow(Base):
    """Redirect old-style "/show/" paths to new-style "/" paths."""

    async def get(self, slug: str) -> None:  # type: ignore
        """Fetch paste from database and redirect to /slug if the paste
        exists."""
        with database.session() as session:
            paste = (
                session.query(database.Paste)
                .filter(database.Paste.slug == slug)
                .first()
            )

            if not paste:
                raise tornado.web.HTTPError(404)

            if paste.exp_date < datetime.utcnow():
                session.delete(paste)
                session.commit()

                log.warn(
                    "RedirectShow.get: paste was expired, is your cronjob running?"
                )

                raise tornado.web.HTTPError(404)

            self.redirect(f"/{paste.slug}")


class FileRaw(Base):
    """Show a file as plaintext."""

    async def get(self, file_id: str) -> None:  # type: ignore
        """Get a file from the database and show it in the plain."""

        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        with database.session() as session:
            file = (
                session.query(database.File)
                .filter(database.File.slug == file_id)
                .first()
            )

            if not file:
                raise tornado.web.HTTPError(404)

            if file.paste.exp_date < datetime.utcnow():
                session.delete(file.paste)
                session.commit()

                log.warn(
                    "FileRaw.get: paste was expired, is your cronjob running?"
                )

                raise tornado.web.HTTPError(404)

            self.set_header("Content-Type", "text/plain; charset=utf-8")
            self.write(file.raw)


class FileHex(Base):
    """Show a file as hexadecimal."""

    async def get(self, file_id: str) -> None:  # type: ignore
        """Get a file from the database and show it in hex."""

        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        with database.session() as session:
            file = (
                session.query(database.File)
                .filter(database.File.slug == file_id)
                .first()
            )

            if not file:
                raise tornado.web.HTTPError(404)

            if file.paste.exp_date < datetime.utcnow():
                session.delete(file.paste)
                session.commit()

                log.warn(
                    "FileRaw.get: paste was expired, is your cronjob running?"
                )

                raise tornado.web.HTTPError(404)

            self.set_header("Content-Type", "text/plain; charset=utf-8")
            self.write(binascii.hexlify(file.raw.encode("utf8")))


class PasteDownload(Base):
    """Download an entire paste."""

    async def get(self, paste_id: str) -> None:  # type: ignore
        """Get all files from the database and download them as a zipfile."""

        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        with database.session() as session:
            paste = (
                session.query(database.Paste)
                .filter(database.Paste.slug == paste_id)
                .first()
            )

            if not paste:
                raise tornado.web.HTTPError(404)

            if paste.exp_date < datetime.utcnow():
                session.delete(paste)
                session.commit()

                log.warn(
                    "FileRaw.get: paste was expired, is your cronjob running?"
                )

                raise tornado.web.HTTPError(404)

            data = io.BytesIO()

            with zipfile.ZipFile(data, "x") as zf:
                for file in paste.files:
                    if file.filename:
                        filename = f"{utility.filename_clean(file.filename)}-{file.slug}.txt"
                    else:
                        filename = f"{file.slug}.txt"

                    zf.writestr(filename, file.raw)

            data.seek(0)

            self.set_header("Content-Type", "application/zip")
            self.set_header(
                "Content-Disposition", f"attachment; filename={paste.slug}.zip"
            )
            self.write(data.read())


class FileDownload(Base):
    """Download a file."""

    async def get(self, file_id: str) -> None:  # type: ignore
        """Get a file from the database and download it in the plain."""

        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        with database.session() as session:
            file = (
                session.query(database.File)
                .filter(database.File.slug == file_id)
                .first()
            )

            if not file:
                raise tornado.web.HTTPError(404)

            if file.paste.exp_date < datetime.utcnow():
                session.delete(file.paste)
                session.commit()

                log.warn(
                    "FileDownload.get: paste was expired, is your cronjob running?"
                )

                raise tornado.web.HTTPError(404)

            self.set_header("Content-Type", "text/plain; charset=utf-8")

            if file.filename:
                filename = (
                    f"{utility.filename_clean(file.filename)}-{file.slug}.txt"
                )
            else:
                filename = f"{file.slug}.txt"

            self.set_header(
                "Content-Disposition", f"attachment; filename={filename}"
            )
            self.write(file.raw)


class Remove(Base):
    """Remove a paste."""

    async def get(self, removal: str) -> None:  # type: ignore
        """Look up if the user visiting this page has the removal id for a
        certain paste. If they do they're authorized to remove the paste."""

        if defensive.ratelimit(self.request, area="delete"):
            raise error.RatelimitError()

        with database.session() as session:
            paste = (
                session.query(database.Paste)
                .filter(database.Paste.removal == removal)
                .first()
            )

            if not paste:
                log.info("RemovePaste.get: someone visited with invalid id")
                raise tornado.web.HTTPError(404)

            if paste.exp_date < datetime.utcnow():
                session.delete(paste)
                session.commit()

                log.warn(
                    "Remove.get: paste was expired, is your cronjob running?"
                )

                raise tornado.web.HTTPError(404)

            session.delete(paste)
            session.commit()

        self.redirect("/")


class RestructuredTextPage(Base):
    """Render a given file as RestructuredText."""

    def initialize(self, file: str) -> None:
        self.file = file

    async def get(self) -> None:
        if defensive.ratelimit(self.request, area="read"):
            raise error.RatelimitError()

        try:
            with open(path.page / self.file) as f:
                html = docutils.core.publish_parts(
                    f.read(), writer_name="html"
                )["html_body"]
        except FileNotFoundError:
            raise tornado.web.HTTPError(404)

        self.render(
            "restructuredtextpage.html",
            html=html,
            pagetitle=(path.page / self.file).stem,
        )


class Logo(Base):
    """Render an image file at the logo path."""

    def initialize(self, path: str) -> None:
        self.path = path

    async def get(self) -> None:
        self.set_header("Content-Type", "image/png")

        with open(self.path, "rb") as f:
            self.write(f.read())



import ipaddress
import logging
import re
from typing import Dict, Union

import token_bucket
from tornado.httputil import HTTPServerRequest

from pinnwand import configuration

log = logging.getLogger(__name__)


ratelimit_area: Dict[
    str,
    Dict[
        Union[ipaddress.IPv4Address, ipaddress.IPv6Address],
        token_bucket.Limiter,
    ],
] = {}


def ratelimit(request: HTTPServerRequest, area: str = "global") -> bool:
    """Test if a requesting IP is ratelimited for a certain area. Areas are
    different functionalities of the website, for example 'view' or 'input' to
    differentiate between creating new pastes (low volume) or high volume
    viewing.

    Important: this does not work well for IPv6 addresses where really when a
    ratelimit is hit, we should walk up the CIDRs (/64, /48, etc) as those
    ranges usually belong to the same person. If this is encountered in real
    life this function will have to be expanded."""

    if area not in ratelimit_area:
        ratelimit_area[area] = {}

    # TODO handle valueerror as validationerror?
    address = ipaddress.ip_address(str(request.remote_ip))

    if address not in ratelimit_area[area]:
        ratelimit_area[area][address] = token_bucket.Limiter(
            configuration.ratelimit[area]["refill"],
            configuration.ratelimit[area]["capacity"],
            token_bucket.MemoryStorage(),
        )

    if not ratelimit_area[area][address].consume(1):
        log.warning("%s hit rate limit for %r", address, area)
        return True

    return False


def spamscore(text: str) -> int:
    """Give a naive spamscore for some text, spam to pinnwand instances seems
    to mostly consist of lists of links so we count the link count vs the total
    text in the paste.

    This will then give a value 0-100, which can be used with a configuration
    option to deny pastes over a certain percentage."""

    # TODO Is this a good URL regex?
    match = re.compile(
        r"[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)"
    )

    text_size = len(text)
    link_size = sum(len(link) for link in match.findall(text))

    score = int((link_size / text_size) * 100)

    log.debug("spamscore: rated at %r score", score)

    return score



database_uri = "sqlite:///:memory:"
paste_size = 256 * 1024  # in bytes
footer = 'View <a href="//github.com/supakeen/pinnwand" target="_BLANK">source code</a>, the <a href="/removal">removal</a> or <a href="/expiry">expiry</a> stories, or read the <a href="/about">about</a> page.'
paste_help = "<p>Welcome to pinnwand, this site is a pastebin. It allows you to share code with others. If you write code in the text area below and press the paste button you will be given a link you can share with others so they can view your code as well.</p><p>People with the link can view your pasted code, only you can remove your paste and it expires automatically. Note that anyone could guess the URI to your paste so don't rely on it being private.</p>"
page_path = None
page_list = ["about", "removal", "expiry"]
preferred_lexers = []  # type: ignore
logo_path = None
report_email = None
expiries = {"1day": 86400, "1week": 604800}
ratelimit = {
    "read": {
        "capacity": 100,
        "consume": 1,
        "refill": 2,
    },
    "create": {
        "capacity": 2,
        "consume": 2,
        "refill": 1,
    },
    "delete": {
        "capacity": 2,
        "consume": 2,
        "refill": 1,
    },
}
spamscore = 50



import logging
import math
import re
from base64 import b32encode
from datetime import datetime
from os import urandom
from typing import Any, Dict, List, Optional

from pygments.lexers import (
    get_all_lexers,
    guess_lexer,
    guess_lexer_for_filename,
)

from pinnwand import database

log = logging.getLogger(__name__)


def list_languages() -> Dict[str, str]:
    # Start with converting the pygments lexers index into a dict.
    lexers = {
        lexer[1][0]: lexer[0]
        for lexer in get_all_lexers()
        # Guard against the lexer format in Pygments not being all that
        # consistent.
        if len(lexer) > 1 and len(lexer[1])
    }

    # Add autodetection option
    lexers["autodetect"] = "Autodetect"

    # Since dicts are sorted since Python 3.7 (and 3.6 per implementation
    # detail) and the Pygments ordering is a bit inane we sort and turn back
    # into a dict here.
    return dict(sorted(lexers.items(), key=lambda x: x[1]))  # type: ignore


GUESS_LANG_OVERRIDES = {"as3": "yaml", "python2": "python"}

GUESS_LANG_IGNORES = ["mime", "tsql"]


def guess_language(raw: str, filename: Optional[str] = None) -> str:
    options = {"stripnl": True}

    # Guess a lexer based on filename and raw text first
    if filename:
        try:
            return str(
                guess_lexer_for_filename(filename, raw, **options).aliases[0]
            )
        except (ValueError, IndexError):
            pass

    # If that didn't work guess lexer just by looking at the raw text
    try:
        language = str(guess_lexer(raw, **options).aliases[0])
    except (ValueError, IndexError):
        # If no lexer was detected, fallback to plain text.
        return "text"

    # These are odd lexers that match far too often, so exclude them.
    if language in GUESS_LANG_IGNORES:
        return "text"

    # Finally check for language overrides and return
    return GUESS_LANG_OVERRIDES.get(language, language)


def hash_create(length: int = 16) -> str:
    return b32encode(urandom(length)).decode("ascii").replace("=", "")


def slug_create(
    auto_scale: bool = True, dont_use: Optional[List[str]] = None
) -> str:
    """Creates a new slug, a slug has to be unique within both the Paste and
    File namespace. These slugs auto-lengthen unless they are specified not
    to."""

    if dont_use is None:
        dont_use = []

    with database.session() as session:
        if auto_scale:
            # We count our new paste as well
            count = (
                session.query(database.Paste).count()
                + session.query(database.File).count()
                + len(dont_use)
                + 1
            )

            # The amount of bits necessary to store that count times two, then
            # converted to bytes with a minimum of 1.

            # We double the count so that we always keep half of the space
            # available (e.g we increase the number of bytes at 127 instead of
            # 255). This ensures that the probing below can find an empty space
            # fast in case of collision.
            necessary = math.ceil(math.log2(count * 2)) // 8 + 1
        else:
            necessary = 16  # 16 bytes should do, right?

        # Now generate random ids in the range with a maximum amount of
        # retries, continuing until an empty slot is found
        tries = 0
        slug = hash_create(necessary)

        # If a slug exists in either the Paste or File namespace create a new
        # one, we also check if we've already generated this slug.
        while any(
            (
                session.query(database.Paste)
                .filter_by(slug=slug)
                .one_or_none(),
                session.query(database.File).filter_by(slug=slug).one_or_none(),
                slug in dont_use,
            )
        ):
            log.debug("slug_create: triggered a collision")
            if tries > 10:
                raise RuntimeError("We exceeded our retry quota on a collision")
            tries += 1
            slug = hash_create(necessary)

        return slug


class SlugContext:
    """Since pinnwand often has to create multiple slugs in one go without
    generating any duplicates we have a context that keeps track of slugs
    already created in the current grouping. See issue #34 for more
    information on the *why*."""

    def __init__(self, auto_scale: bool = True) -> None:
        self._slugs: List[str] = []
        self._auto_scale = auto_scale

    def __enter__(self) -> "SlugContext":
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return None

    def __next__(self) -> str:
        slug = slug_create(
            self._auto_scale,
            self._slugs,
        )
        self._slugs.append(slug)
        return slug


units = [
    (1 << 20, "Mb"),
    (1 << 10, "kb"),
    (1, "b"),
]


def size_postfix(count: int) -> str:
    for f, p in units:
        if count > f:
            break

    return f"{count/f:.0f}{p}"


def filename_clean(filename: str) -> str:
    """Try to clean a filename for safe consumption as much as we can. These
    filenames are sent to other users."""

    # If there's a dot we remove the suffix
    if "." in filename:
        filename = filename.rsplit(".", 1)[0]

    # Replace anything that could be misinterpreted by an operating system
    return re.sub(r"[^A-Za-z0-9-_]", "", filename)


def reap() -> None:
    """Delete all pastes that are past their expiry date in pinnwand's
    database."""

    with database.session() as session:
        pastes = (
            session.query(database.Paste)
            .filter(database.Paste.exp_date < datetime.utcnow())
            .all()
        )

        for paste in pastes:
            session.delete(paste)

        session.commit()

        log.info("reap: removed %d pastes", len(pastes))


async def async_reap() -> None:
    """Just the asynchronous definition of `reap`, note: not async."""
    reap()



"""Collection of pinnwand's command line entry points that allow you to start
a HTTP server, add and remove paste, initialize the database and reap expired
pastes."""

import ast
import logging
import os
import sys
from datetime import timedelta
from typing import TYPE_CHECKING, Optional

import click
import tornado.ioloop

log = logging.getLogger(__name__)


@click.group()
@click.option(
    "--verbose",
    "-v",
    count=True,
    help="Verbosity, passing more heightens the verbosity.",
)
@click.option("--configuration-path", default=None, help="Configuration file.")
def main(verbose: int, configuration_path: Optional[str]) -> None:
    """Pinnwand pastebin software."""
    logging.basicConfig(level=10 + (logging.FATAL - verbose * 10))

    from pinnwand import configuration

    # First check if we have a configuration path
    if configuration_path:
        if (
            TYPE_CHECKING
        ):  # lie to mypy, see https://github.com/python/mypy/issues/1153
            import tomllib as toml
        else:
            try:
                import tomllib as toml
            except ImportError:
                import tomli as toml

        with open(configuration_path, "rb") as file:
            configuration_file = toml.load(file)

            for key, value in configuration_file.items():
                setattr(configuration, key, value)

    # Or perhaps we have configuration in the environment, these are prefixed
    # with PINNWAND_ and all upper case, remove the prefix, convert to
    # lowercase.
    for key, value in os.environ.items():
        if key.startswith("PINNWAND_"):
            key = key.removeprefix("PINNWAND_")
            key = key.lower()

            try:
                value = ast.literal_eval(value)
            except ValueError:
                # When `ast.literal_eval` can't parse the value into another
                # type we take it at string value
                pass

            setattr(configuration, key, value)

    from pinnwand import database

    database.Base.metadata.create_all(database._engine)


@main.command()
@click.option("--port", default=8000, help="Port to listen to.")
def http(port: int) -> None:
    """Run pinnwand's HTTP server."""
    from pinnwand import utility
    from pinnwand.http import make_application

    # Reap expired pastes on startup (we might've been shut down for a while)
    utility.reap()

    # Schedule reaping every 1800 seconds
    reap_task = tornado.ioloop.PeriodicCallback(utility.async_reap, 1_800_000)
    reap_task.start()

    application = make_application()
    application.listen(port, xheaders=True)

    tornado.ioloop.IOLoop.current().start()


@main.command()
@click.option("--lexer", default="text", help="Lexer to use.")
def add(lexer: str) -> None:
    """Add a paste to pinnwand's database from stdin."""
    from pinnwand import database, utility

    if lexer not in utility.list_languages():
        log.error("add: unknown lexer")
        return

    paste = database.Paste(
        utility.slug_create(), expiry=timedelta(days=1).seconds
    )
    file = database.File(paste.slug, sys.stdin.read(), lexer=lexer)
    paste.files.append(file)

    with database.session() as session:
        session.add(paste)
        session.commit()

        log.info("add: paste created: %s", paste.slug)


@main.command()
@click.option("--paste", help="database.Paste identifier.", required=True)
def delete(paste: str) -> None:
    """Delete a paste from pinnwand's database."""
    from pinnwand import database

    with database.session() as session:
        paste_object = (
            session.query(database.Paste)
            .filter(database.Paste.slug == paste)
            .first()
        )

        if not paste_object:
            log.error("delete: unknown paste")
            raise SystemExit(1)

        session.delete(paste_object)
        session.commit()

        log.info("delete: paste %s deleted", paste_object)


@main.command()
def reap() -> None:
    """Delete all pastes that are past their expiry date in pinnwand's
    database."""
    from pinnwand import utility

    log.warning("reap: this command is deprecated and will be removed in 1.6")

    utility.reap()


@main.command()
def resyntax() -> None:
    """Rerun `pygments` over all files in the database to update their formatted
    output."""
    import pygments.lexers
    from pygments_better_html import BetterHtmlFormatter

    from pinnwand import database, utility

    with database.session() as session:
        files = session.query(database.File).all()

        for file in files:
            if file.lexer == "autodetect":
                lexer = utility.guess_language(file.raw, file.filename)
                log.debug(f"resyntax: Language guessed as {lexer}")

            lexer = pygments.lexers.get_lexer_by_name(lexer)
            formatter = BetterHtmlFormatter(  # pylint: disable=no-member
                linenos="table", cssclass="source"
            )

            formatted = pygments.highlight(file.raw, lexer, formatter)

            file.fmt = formatted

        session.commit()

        log.info("resyntax: highlighted %d pastes", len(files))



import logging
import secrets
from typing import Any, List

import tornado.web

from pinnwand import configuration, handler, path

log = logging.getLogger(__name__)


def make_application() -> tornado.web.Application:
    pages: List[Any] = [
        (r"/", handler.website.Create),
        (r"/\+(.*)", handler.website.Create),
        (r"/create", handler.website.CreateAction),
        (r"/show/([A-Z2-7]+)(?:#.+)?", handler.website.RedirectShow),
        (r"/repaste/([A-Z2-7]+)(?:#.+)?", handler.website.Repaste),
        (r"/raw/([A-Z2-7]+)(?:#.+)?", handler.website.FileRaw),
        (r"/([A-Z2-7]+)(?:#.+)?/raw", handler.website.FileRaw),
        (r"/hex/([A-Z2-7]+)(?:#.+)?", handler.website.FileHex),
        (r"/([A-Z2-7]+)(?:#.+)?/hex", handler.website.FileHex),
        (r"/download/([A-Z2-7]+)(?:#.+)?", handler.website.FileDownload),
        (r"/([A-Z2-7]+)(?:#.+)?/download", handler.website.FileDownload),
        (
            r"/download-archive/([A-Z2-7]+)(?:#.+)?",
            handler.website.PasteDownload,
        ),
        (
            r"/([A-Z2-7]+)(?:#.+)?/download-archive",
            handler.website.PasteDownload,
        ),
        (r"/remove/([A-Z2-7]+)", handler.website.Remove),
    ]

    pages += [
        (
            f"/{file}",
            handler.website.RestructuredTextPage,
            {"file": f"{file}.rst"},
        )
        for file in configuration.page_list
    ]

    pages += [
        (r"/api/v1/paste", handler.api_v1.Paste),
        (r"/api/v1/lexer", handler.api_v1.Lexer),
        (r"/api/v1/expiry", handler.api_v1.Expiry),
        (r"/json/new", handler.api_deprecated.Create),
        (r"/json/remove", handler.api_deprecated.Remove),
        (r"/json/show/([A-Z2-7]+)(?:#.+)?", handler.api_deprecated.Show),
        (r"/json/lexers", handler.api_deprecated.Lexer),
        (r"/json/expiries", handler.api_deprecated.Expiry),
        (r"/curl", handler.api_curl.Create),
    ]

    if configuration.logo_path:
        pages += [
            (
                r"/static/logo.png",
                handler.website.Logo,
                {"path": configuration.logo_path},
            ),
            (
                r"/static/favicon.png",
                handler.website.Logo,
                {"path": configuration.logo_path},
            ),
        ]

    pages += [
        (
            r"/static/(.*)",
            tornado.web.StaticFileHandler,
            {"path": path.static},
        ),
        (r"/(.*)(?:#.+)?", handler.website.Show),
    ]

    app = tornado.web.Application(
        pages,
        template_path=path.template,
        default_handler_class=handler.website.Base,
        xsrf_cookies=True,
        cookie_secret=secrets.token_hex(),
        static_path=path.static,
        xheaders=True,
    )

    app.configuration = configuration  # type: ignore

    return app



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pinnwand', 'pinnwand.handler']

package_data = \
{'': ['*'], 'pinnwand': ['page/*', 'static/*', 'template/*', 'template/part/*']}

install_requires = \
['click>=8.1,<9.0',
 'docutils>=0.19,<0.20',
 'pygments-better-html>=0.1.4,<0.2.0',
 'pygments>=2.13,<3.0',
 'sqlalchemy>=1.4,<2.0',
 'token-bucket>=0.3.0,<0.4.0',
 'tornado>=6.2,<7.0']

extras_require = \
{':python_version < "3.11"': ['tomli>=2.0.1,<3.0.0']}

entry_points = \
{'console_scripts': ['pinnwand = pinnwand.__main__:main']}

setup_kwargs = {
    'name': 'pinnwand',
    'version': '1.4.0',
    'description': 'Straightforward pastebin software.',
    'long_description': '![pinnwand logo, a rabbit](https://pinnwand.readthedocs.io/en/latest/_static/logo-doc.png)\n\n# pinnwand\n\n[![](https://travis-ci.org/supakeen/pinnwand.svg?branch=master)](https://travis-ci.org/supakeen/pinnwand) [![](https://readthedocs.org/projects/pinnwand/badge/?version=latest)](https://pinnwand.readthedocs.io/en/latest/) [![](https://pinnwand.readthedocs.io/en/latest/_static/license.svg)](https://github.com/supakeen/pinnwand/blob/master/LICENSE) [![](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black) [![](https://img.shields.io/pypi/v/pinnwand)](https://pypi.org/project/pinnwand) [![](https://codecov.io/gh/supakeen/pinnwand/branch/master/graph/badge.svg)](https://codecov.io/gh/supakeen/pinnwand) [![](https://quay.io/repository/supakeen/pinnwand/status)](https://quay.io/repository/supakeen/pinnwand)\n\n## About\n\n`pinnwand` is Python pastebin software that tried to keep it simple but got\na little more complex.\n\nPrerequisites\n=============\n* Python >= 3.7\n* Tornado\n* sqlalchemy\n* click\n* docutils\n* tomli\n* pygments-better-html\n* a database driver\n\nUsage\n=====\n\nWeb\n---\nEnter text, click "Paste", easy enough.\n\nsteck\n-----\n[steck](https://supakeen.com/project/steck) is a command line client to pinnwand instances:\n\n```\n€ pip install --user steck\n...\n€ steck paste *\nYou are about to paste the following 7 files. Do you want to continue?\n- LICENSE\n- mypy.ini\n- poetry.lock\n- pyproject.toml\n- README.rst\n- requirements.txt\n- steck.py\n\nContinue? [y/N] y\n\nCompleted paste.\nView link:    https://localhost:8000/W5\nRemoval link: https://localhost:8000/remove/TS2AFFIEHEWUBUV5HLKNAUZFEI\n```\n\ncurl\n----\n`pinnwand` has a direct endpoint for `curl` users:\n\n```\n€ echo "foo" | curl -X POST http://localhost:8000/curl -F \'raw=<-\'\nPaste URL:   http://localhost:8000/OE\nRaw URL:     http://localhost:8000/raw/GU\nRemoval URL: http://localhost:8000/remove/GQBHGJYKRWIS34D6FNU6CJ3B5M\n€ curl http://localhost:8000/raw/GU\nfoo%\n```\n\nThis will preselect the `lexer` and `expiry` arguments to be `text` and\n`1day` respectively. You can provide those to change them.\n\nAPI\n---\n`pinnwand` provides a straight forward JSON API, here\'s an example using the\ncommon requests library:\n\n```\n>>> requests.post(\n...     "http://localhost:8000/api/v1/paste",\n...     json={\n...             "expiry": "1day",\n...             "files": [\n...                     {"name": "spam", "lexer": "python", "content": "eggs"},\n...             ],\n...     }\n... ).json()\n{\'link\': \'http://localhost:8000/74\', \'removal\': \'http://localhost:8000/remove/KYXQLPZQEWV2L4YZM7NYGTR7TY\'}\n```\n\nMore information about this API is available in the [documentation](https://pinnwand.readthedocs.io/en/latest/).\n\n\nMore ways to use pinnwand\n-------------------------\nVarious deprecated ways of posting are still supported, don\'t implement these\nfor any new software but if you are maintaining old software and want to know\nhow they used to work you can read our documentation_.\n\nIf you do use a deprecated endpoint to post a warning will be shown below any\npastes that are created this way.\n\nReporting bugs\n==============\nBugs are reported best at `pinnwand`\'s [project page](https://github.com/supakeen/pinnwand) on github. If you just\nwant to hang out and chat about `pinnwand` then I\'m available in the\n`#pinnwand` channel on Freenode IRC.\n\nLicense\n=======\n`pinnwand` is distributed under the MIT license. See `LICENSE`\nfor details.\n\nHistory\n=======\nThis pastebin has quite a long history which isn\'t reflected entirely in its\nrepository.\n',
    'author': 'Simon de Vlieger',
    'author_email': 'cmdr@supakeen.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/supakeen/pinnwand',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



class ValidationError(ValueError):
    """This exception is used to indicate that a certain requst is lacking or
    has unacceptable data aboard."""

    pass


class RatelimitError(ValueError):
    """This exception is used to indicate that a user has surpassed their
    ratelimit for a certain action."""

    pass


class SpamError(ValueError):
    """This exception is used to indicate that a user has exceeded the allowed
    spamscore."""

    pass



import logging
from urllib.parse import urljoin

import tornado.web

from pinnwand import configuration, database, defensive, utility

log = logging.getLogger(__name__)


class Create(tornado.web.RequestHandler):
    def check_xsrf_cookie(self) -> None:
        return

    def get(self) -> None:
        raise tornado.web.HTTPError(400)

    def post(self) -> None:
        if defensive.ratelimit(self.request, area="create"):
            self.set_status(429)
            self.write("Enhance your calm, you have exceeded the ratelimit.")
            return

        lexer = self.get_body_argument("lexer", "text")
        raw = self.get_body_argument("raw", "", strip=False)
        expiry = self.get_body_argument("expiry", "1day")

        self.set_header("Content-Type", "text/plain")

        if lexer not in utility.list_languages():
            log.info(
                "CurlCreate.post: a paste was submitted with an invalid lexer"
            )
            self.set_status(400)
            self.write("Invalid `lexer` supplied.\n")
            return

        # Guard against empty strings
        if not raw or not raw.strip():
            log.info("CurlCreate.post: a paste was submitted without raw")
            self.set_status(400)
            self.write("Invalid `raw` supplied.\n")
            return

        if expiry not in configuration.expiries:
            log.info("CurlCreate.post: a paste was submitted without raw")
            self.set_status(400)
            self.write("Invalid `expiry` supplied.\n")
            return

        paste = database.Paste(
            utility.slug_create(), configuration.expiries[expiry], "curl"
        )
        file = database.File(paste.slug, raw, lexer)
        paste.files.append(file)

        with database.session() as session:
            session.add(paste)
            session.commit()

            # The removal cookie is set for the specific path of the paste it is
            # related to
            self.set_cookie(
                "removal", str(paste.removal), path=f"/{paste.slug}"
            )

            url_request = self.request.full_url()
            url_paste = urljoin(url_request, f"/{paste.slug}")
            url_removal = urljoin(url_request, f"/remove/{paste.removal}")
            url_raw = urljoin(url_request, f"/raw/{file.slug}")

            self.write(
                f"Paste URL:   {url_paste}\nRaw URL:     {url_raw}\nRemoval URL: {url_removal}\n"
            )
